/* Copyright (C) 2006 P.L. Lucas
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA. 
 */

#include <QBoxLayout>
#include <QToolBar>
#include <QIcon>
#include <QWorkspace>
#include <QFileDialog>
#include <QMessageBox>
#include <QTextCursor>
#include "editor.h"
#include "config.h"
#include "search_dialog.h"
#include "numberedtextview.h"

Editor * Editor::instance = 0;

// Public Static Constructor
// It allows to create only one instance of Editor simultaneously
Editor * Editor::createEditor(QWidget * parent, bool multiple) {

	if (instance == 0 || multiple) {
		instance = new Editor(parent);
		return instance;
	} else {
		return 0;
	} 
}

// Private Constructor
Editor::Editor(QWidget *parent): BaseWidget(parent)
{
  currentNtv=NULL;
  search_dialog=NULL;
  // Title
  setWindowTitle(tr("Editor"));
  setWindowIcon(QIcon(QString(ICON_PATH) + "/xedit.png"));

  // Layout
  QVBoxLayout *vLayout = new QVBoxLayout;
  setLayout(vLayout);

  resize(500, 400);

  // Toolbar
  toolBar = new QToolBar(this);
  toolBar->setOrientation(Qt::Horizontal);
  toolBar->setMinimumHeight(40);
  toolBar->setMaximumHeight(40);
  connect(toolBar, SIGNAL(actionTriggered(QAction*)),
	  this, SLOT(toolbar_action(QAction*)));
  vLayout->addWidget(toolBar);
  toolBar->show();

  // Toolbar buttons
  actionNew = toolBar->addAction(QIcon(QString(ICON_PATH) + "/filenew.png"), tr("New (Ctrl+N)"));
  actionNew->setShortcut(tr("Ctrl+N"));
  actionOpen = toolBar->addAction(QIcon(QString(ICON_PATH) + "/fileopen"), tr("Open (Ctrl+O)"));
  actionOpen->setShortcut(tr("Ctrl+O"));
  actionSave = toolBar->addAction(QIcon(QString(ICON_PATH) + "/filesave"), tr("Save (Ctrl+S)"));
  actionSave->setShortcut(tr("Ctrl+S"));
  actionSaveAs = toolBar->addAction(QIcon(QString(ICON_PATH) + "/filesaveas.png"), tr("Save as"));
  actionClose = toolBar->addAction(QIcon(QString(ICON_PATH) + "/button_cancel.png"), tr("Close"));
  toolBar->addSeparator();

  actionRun = toolBar->addAction(QIcon(QString(ICON_PATH) + "/run.png"), tr("Run"));
  actionStep = toolBar->addAction(QIcon(QString(ICON_PATH) + "/step.png"), tr("Next line"));
  actionStep->setEnabled(false);
  actionSendToOctave=toolBar->addAction(QIcon(QString(ICON_PATH) + "/konsole.png"), tr("Send to Octave") );
  toolBar->addSeparator();

  actionUndo = toolBar->addAction(QIcon(QString(ICON_PATH) + "/undo.png"), tr("Undo (Ctrl+Z)"));
  actionUndo->setShortcut(tr("Ctrl+Z"));
  actionCut = toolBar->addAction(QIcon(QString(ICON_PATH) + "/editcut"), tr("Cut (Ctrl+X)"));
  actionCut->setShortcut(tr("Ctrl+X"));
  actionCopy = toolBar->addAction(QIcon(QString(ICON_PATH) + "/editcopy"), tr("Copy (Ctrl+C)"));
  actionCopy->setShortcut(tr("Ctrl+C"));
  actionPaste = toolBar->addAction(QIcon(QString(ICON_PATH) + "/editpaste"),  tr("Paste (Ctrl+V)"));
  actionPaste->setShortcut(tr("Ctrl+V"));

  toolBar->addSeparator();

  actionSearch = toolBar->addAction(QIcon(QString(ICON_PATH) + "/search.png"),  tr("Search and replace"));

  // TabWidget
  tabWidget = new QTabWidget(this);
  tabWidget->show();

  connect(tabWidget, SIGNAL(currentChanged(int)),
	  this, SLOT(tabChanged(int)));

  vLayout->addWidget(tabWidget);

  toolbar_action(actionNew);

  setAcceptDrops(true);
}

Editor::~Editor() {
	instance = 0;
	// if(search_dialog!=NULL) delete search_dialog;
}


void Editor::dragEnterEvent(QDragEnterEvent *event)
{
	if (event->mimeData()->hasFormat("text/plain"))
		event->acceptProposedAction();
}


void Editor::dropEvent ( QDropEvent * event )
{
	QString path=event->mimeData()->text();

      try{
	// Si la pestaña activa no contiene ningún texto ni es un archivo,
	// abrir en ella
	if(currentNtv->path().isEmpty() &&
	   currentNtv->textEdit()->document()->isEmpty())
	{
	  tabWidget->setTabText(tabWidget->currentIndex(), path.split("/").last());

	}else
	{
	  CodeEdit *codeEdit = new CodeEdit(NULL, QString(CONFIG_PATH) + "/octave.xml");
	  NumberedTextView *ntv = new NumberedTextView(NULL, codeEdit);
	  connect(ntv->textEdit(), SIGNAL(toggleBreakpoint(int)), this, SLOT(toggleBreakpoint(int)));
	  connect(ntv, SIGNAL(textModified()), this, SLOT(textModified()));

	  tabWidget->setCurrentIndex(tabWidget->addTab(ntv, path.split("/").last()));
	}  

	currentNtv->open(path);

      }catch(...){
	QMessageBox::critical(NULL, tr("Error"), path + " can not be opened");
      }
	
	event->acceptProposedAction();
}


void Editor::setOctaveConnection(OctaveConnection *oc)
{
  octave_connection = oc;
}


void Editor::toolbar_action(QAction *action)
{
  QStringList filters;

  filters << "Octave (*.m; *.M)"
	  << "Plain text (*.txt)"
	  << "All files (*)";

  
  if(action == actionNew){
    /** New **/
    CodeEdit *codeEdit = new CodeEdit(NULL, QString(CONFIG_PATH) + "/octave.xml");
    NumberedTextView *ntv = new NumberedTextView(NULL, codeEdit);
    connect(ntv->textEdit(), SIGNAL(toggleBreakpoint(int)), this, SLOT(toggleBreakpoint(int)));
    connect(ntv, SIGNAL(textModified()), this, SLOT(textModified()));

    tabWidget->setCurrentIndex(tabWidget->addTab(ntv, tr("New")));

    currentNtv = ntv;

  }else if(action == actionOpen){
    /** Open **/
    QString path;
    QFileDialog openDialog(this, Qt::Dialog);

    openDialog.setAcceptMode(QFileDialog::AcceptOpen);
    openDialog.setDefaultSuffix("m");
    openDialog.setFilters(filters);

    if(openDialog.exec() == QDialog::Accepted)
    {
      path = openDialog.selectedFiles().first();

      try{
	// Si la pestaña activa no contiene ningún texto ni es un archivo,
	// abrir en ella
	if(currentNtv->path().isEmpty() &&
	   currentNtv->textEdit()->document()->isEmpty())
	{
	  tabWidget->setTabText(tabWidget->currentIndex(), path.split("/").last());

	}else
	{
	  CodeEdit *codeEdit = new CodeEdit(NULL, QString(CONFIG_PATH) + "/octave.xml");
	  NumberedTextView *ntv = new NumberedTextView(NULL, codeEdit);
	  connect(ntv->textEdit(), SIGNAL(toggleBreakpoint(int)), this, SLOT(toggleBreakpoint(int)));
	  //connect(ntv, SIGNAL(textModified()), this, SLOT(textModified()));

	  tabWidget->setCurrentIndex(tabWidget->addTab(ntv, path.split("/").last()));
	}  

	disconnect(currentNtv, SIGNAL(textModified()), this, SLOT(textModified()));
	currentNtv->open(path);
	connect(currentNtv, SIGNAL(textModified()), this, SLOT(textModified()));

      }catch(...){
	QMessageBox::critical(NULL, tr("Error"), path + " can not be opened");
      }
    }

  }else if(action == actionSave && !currentNtv->path().isEmpty()){  
    /* Save */
    try{
      currentNtv->save();
      tabWidget->setTabText(tabWidget->currentIndex(), currentNtv->path().split("/").last());
    }catch(...){
      QMessageBox::critical(NULL, tr("Error"), "Can not be saved");
    }

  }else if(action == actionSaveAs || (action == actionSave && currentNtv->path().isEmpty())){
    /** Save as **/
    QString path;
    QFileDialog saveDialog(this, Qt::Dialog);

    saveDialog.setAcceptMode(QFileDialog::AcceptSave);
    saveDialog.setDefaultSuffix("m");
    saveDialog.setFilters(filters);

    if(saveDialog.exec() == QDialog::Accepted)
    {
      path = saveDialog.selectedFiles().first();

      try{
	currentNtv->save(path);
	tabWidget->setTabText(tabWidget->currentIndex(), path.split("/").last());
      }catch(...){
	QMessageBox::critical(NULL, tr("Error"), path + "can not be saved");
      }
    }

  }else if(action == actionRun){
    /** Run */
    if(currentNtv->path().isEmpty())
    {
      QMessageBox::critical(NULL, tr("Error"), tr("You must save the file first"));
      return;
    }

    // Debug?
    if(actionStep->isEnabled())
      octave_connection->command_enter(QString("dbcont"));
    else
    {
      QFileInfo finfo(currentNtv->path());
      QList<int> *breakpoints = currentNtv->getBreakpoints();
      
      // Source
      octave_connection->command_enter(QString("source('") + finfo.absoluteFilePath() + "')");

      // Clear breakpoints
      octave_connection->command_enter(QString("dbclear('") + finfo.baseName() + "','')");

      // Insert breakpoints
      for(QList<int>::const_iterator i = breakpoints->constBegin();
	  i != breakpoints->constEnd();
	  i++)
	{
	  octave_connection->command_enter(QString("dbstop('")
					   + finfo.baseName()
					   + "','" + QString::number(*i) + "')");
	}

      // Connect debug
      connect(octave_connection, SIGNAL(debug(int, int)),
	      this, SLOT(debug(int, int)));
      connect(octave_connection, SIGNAL(endDebug()),
	      this, SLOT(endDebug()));

      // Run
      octave_connection->command_enter(finfo.baseName());
    } // End debug?

  }else if(action == actionUndo){
    // Undo
    currentNtv->textEdit()->document()->undo();
  }else if(action == actionCut){
    // Cut
    currentNtv->textEdit()->cut();
  }else if(action == actionCopy){
    // Copy
    currentNtv->textEdit()->copy();
  }else if(action == actionPaste){
    // Paste
    currentNtv->textEdit()->paste();
  }else if(action == actionSearch){
	if(search_dialog==NULL)
	{
		search_dialog=new SearchDialog(this);
		connect(search_dialog, SIGNAL(search_signal()), this, SLOT(search()));
		connect(search_dialog, SIGNAL(replace_signal()), this, SLOT(replace()));
	}
	search_dialog->show();
  }else if(action == actionSendToOctave){
    octave_connection->command_enter( currentNtv->textEdit()->document()->toPlainText() );
  }else if(action == actionStep){
    octave_connection->command_enter( "dbstep" );
  }else if(action == actionClose){
    // ¿Modificado?
    if(currentNtv->modified())
    {
      QMessageBox msg(tr("Close"), tr("The file has been modified. Save changes?"),
		      QMessageBox::Question,
		      QMessageBox::Yes, QMessageBox::No,
		      QMessageBox::Cancel | QMessageBox::Default,
		      this);

      switch(msg.exec())
      {
        case QMessageBox::Yes:
	  toolbar_action(actionSave);
	  break;
        case QMessageBox::No:
	  // No hacer nada
	  break;
        default:
	  return;
      }
    }

    // Borrar
    //tabWidget->removeTab(tabWidget->currentIndex());
    printf("Se borra el actual\n");
    delete currentNtv;
    printf("Se pone a NULL\n");

    // Crear uno si no queda ninguno
    if(tabWidget->count() == 0)
    {
      CodeEdit *codeEdit = new CodeEdit(NULL, QString(CONFIG_PATH) + "/octave.xml");
      NumberedTextView *ntv = new NumberedTextView(NULL, codeEdit);
      connect(ntv->textEdit(), SIGNAL(toggleBreakpoint(int)), this, SLOT(toggleBreakpoint(int)));
      connect(ntv, SIGNAL(textModified()), this, SLOT(textModified()));

      tabWidget->addTab(ntv, tr("New"));

      currentNtv = ntv;
    }
    else
    {
    	//tabWidget->setCurrentIndex(tabWidget->currentIndex());
    	tabChanged(tabWidget->currentIndex());
    }
  }else{
    printf("Unhandled action\n");
  }
}

void Editor::search()
{
  QString search, replace;
  QTextCursor cursor;
  QTextEdit *textEdit = currentNtv->textEdit();

  // Strings
  search = search_dialog->searchString();
  replace = search_dialog->replaceString();

  // Flags
  QTextDocument::FindFlags flags;
  if(search_dialog->caseSensitive())
    flags |= QTextDocument::FindCaseSensitively;
  if(search_dialog->wholeWords())
    flags |= QTextDocument::FindWholeWords;

  // Search
  cursor = textEdit->textCursor();
  if(search_dialog->searchStringIsRegExp())
  {
    // Search string is a regular expression
    QRegExp searchReg(search);
    
    //cursor = textEdit->document()->find(searchReg, cursor, flags);
    cursor = textEdit->document()->find(search, cursor, flags);
  }else{
    // Search string is not a regular expression
    cursor = textEdit->document()->find(search, cursor, flags);
  }

  textEdit->setTextCursor(cursor);
}

void Editor::replace()
{
  QTextCursor cursor = currentNtv->textEdit()->textCursor();

  if(!cursor.hasSelection())
    cursor.insertText(search_dialog->replaceString());

  search();
}

void Editor::toggleBreakpoint(int lineno)
{
  currentNtv->toggleBreakpoint(lineno);
}

void Editor::debug(int lineno, int /*colno*/)
{
  currentNtv->setCurrentLine(lineno);
  actionStep->setEnabled(true);
}

void Editor::endDebug()
{
  currentNtv->setCurrentLine(-1);
  actionStep->setEnabled(false);
}

void Editor::tabChanged(int index)
{
  //printf("Activado %d\n", index);
  //if(currentNtv!=NULL)
  //	disconnect(currentNtv->textEdit(), SIGNAL(toggleBreakpoint(int)), this, SLOT(toggleBreakpoint(int)));
  currentNtv = (NumberedTextView*)tabWidget->widget(index);
  setWindowTitle(tabWidget->tabText(index));
  //connect(currentNtv->textEdit(), SIGNAL(toggleBreakpoint(int)), this, SLOT(toggleBreakpoint(int)));
}

void Editor::textModified()
{
	if(currentNtv->path().isEmpty())
	{
		tabWidget->setTabText(tabWidget->currentIndex(), tr("New")+"*");
	}
	else
	{
		tabWidget->setTabText(tabWidget->currentIndex(), currentNtv->path().split("/").last()+"*");
	}
}


void Editor::closeEvent ( QCloseEvent * event )
{
	bool modified=false;
	for(int i=0;i<tabWidget->count();i++)
	{
		modified|=( (NumberedTextView*)tabWidget->widget(i) )->modified();
	}
	int ok;
	if(modified)
		ok=QMessageBox::warning (this,tr("Close this window?"), tr("You are going to close Editor. Are you sure?"), QMessageBox::Ok, QMessageBox::Cancel);
	else
		ok=QMessageBox::Ok;

	if (ok==QMessageBox::Ok) {
		event->accept();
	} else {
		event->ignore();
	}
}

